@@ -10,12 +10,9 @@ import uuid
 import traceback
 
 logger = logging.getLogger(__name__)
-logging.basicConfig()
-logger.propagate = False
 
 class Client:
     def __init__(self, local_name, loop = None, port=13254):
-        logger.setLevel(logging.DEBUG)
         self.uri = f"ws://localhost:{port}"        
         self.local_name = local_name
         self.websocket = None
@@ -24,12 +21,15 @@ class Client:
         
         self.loop = loop or asyncio.get_running_loop()
         self.authorized = False
+        self.on_hold = False
         self.events = {
             "on_winerp_connect": None,
             "on_winerp_ready": None,
             "on_winerp_disconnect": None,
             "on_winerp_request": None,
-            "on_winerp_response": None
+            "on_winerp_response": None,
+            "on_winerp_information": None,
+            "on_winerp_error": None
         }
     
     async def __empty_event(self, *args, **kwargs):
@@ -52,7 +52,7 @@ class Client:
     
     def __send_message(self, data):
         self.loop.create_task(self.send_message(data))
-        
+    
     async def start(self):
         if self.websocket is None or self.websocket.closed:
             logger.info("Connecting to Websocket")
@@ -84,6 +84,14 @@ class Client:
             
         return route_decorator
     
+    async def ping(self):
+        if self.on_hold or self.websocket is None or not self.websocket.open:
+            raise ClientNotReadyError("The client is currently not ready to send or accept requests.")
+        if not self.authorized:
+            raise UnauthorizedError("Client is not authorized!")
+
+        await self.send_message(MessagePayload(type=Payloads.ping))
+
     def get_response(
         self,
         _uuid: str,
@@ -111,10 +119,15 @@ class Client:
             `asyncio.TimeoutError`: If the response is not received within the timeout.
         '''
         if self.websocket is not None and self.websocket.open:
+            if self.on_hold:
+                raise ClientNotReadyError("The client is currently not ready to send or accept requests.")
             if not self.authorized:
                 raise UnauthorizedError("Client is not authorized!")
             
-            logger.info("Requesting IPC Server for %r with data %r", route, kwargs)
+            if not route or not source:
+                raise ValueError("Missing required information for this request")
+
+            logger.info("Requesting IPC Server for %r", route)
         
             _uuid = str(uuid.uuid4())
             payload = MessagePayload(
@@ -129,6 +142,37 @@ class Client:
             await self.send_message(payload)
             recv = await self.get_response(_uuid, self.loop, timeout=timeout)
             return recv
+        
+        else:
+            raise ClientNotReadyError("The client has not been started or has disconnected")
+
+    async def inform(
+        self,
+        data: any,
+        destinations: list
+    ):
+
+        if self.websocket is not None and self.websocket.open:
+            if self.on_hold:
+                raise ClientNotReadyError("The client is currently not ready to send or accept requests.")
+            if not self.authorized:
+                raise UnauthorizedError("Client is not authorized!")
+
+            logger.info("Informing IPC Server to redirect to routes %s" % destinations)
+            if not isinstance(destinations, list):
+                destinations = [destinations]
+
+            payload = MessagePayload(
+                type = Payloads.information,
+                id = self.local_name,
+                route = destinations,
+                data = data,
+            )
+
+            await self.send_message(payload)
+        else:
+            raise ClientNotReadyError("The client has not been started or has disconnected")
+
 
     async def __on_message(self):
         while True:
@@ -136,11 +180,13 @@ class Client:
                 message = WsMessage(json.loads(await self.websocket.recv()))
             except websockets.exceptions.ConnectionClosedError:
                 self.loop.create_task(self.get_event("on_winerp_disconnect")())
+                break
 
             if message.type.success and not self.authorized:
                 logger.info("Authorized Successfully")
                 self.loop.create_task(self.get_event("on_winerp_ready")())
                 self.authorized = True
+                self.on_hold = False
 
             elif message.type.request:
                 if message.route not in self.routes:
@@ -165,9 +211,22 @@ class Client:
                 self.loop.create_task(self.get_event("on_winerp_response")())
 
             elif message.type.error:
-                logger.warning("Failed to fulfill request: %s" % message.error)
+                if message.data == "Already authorized.":
+                    self.on_hold = True
+                    logger.warn("Another client is already connected. Requests will be enabled when the other is disconnected.")
+                else:
+                    logger.debug("Failed to fulfill request: %s" % message.data)
+                    self.loop.create_task(self.get_event("on_winerp_error")(message.data))
+
                 if message.uuid is not None:
                     self.loop.create_task(self._dispatch(message))
+
+            elif message.type.information:
+                if message.data:
+                    logger.info("Received an information bit from client: %s" % message.id)
+                    self.loop.create_task(self.get_event("on_winerp_information")(message.data, message.id))
+                
+
     
     async def _fulfill_request(self, message: WsMessage):
         route = message.route
@@ -180,8 +239,10 @@ class Client:
 
         try:
             payload.data = await func(**data)
+            json.dumps(payload.data) #Ensuring data is serializable
         except Exception as error:
             logger.exception(error)
+            self.loop.create_task(self.get_event("on_winerp_error")(error))
             etype = type(error)
             trace = error.__traceback__
             lines = traceback.format_exception(etype, error, trace)
